import torch 
torch.manual_seed(42)
torch.set_num_threads(1)


def einsum(T, U):
    return torch.einsum('axob,amz,moyn,bnw->xwzy',(T, U, T, U))

def tensordot(T, U):
    TU1 = torch.tensordot(T, U, dims=([0], [0]))
    TU2 = torch.tensordot(T, U, dims=([3], [1]))
    T = torch.tensordot(TU1, TU2, dims=([1,2,3], [1,3,0]))
    T = T.permute(0,3,1,2).contiguous()
    return T

if __name__=='__main__':
    import time
    dtype = torch.float64
    device = 'cuda:5'
    
    D1 = 30
    D2 = 30
    chi = 30
    T = torch.randn(D1, D2, D2, D1, device=device, dtype=dtype)
    U = torch.randn(D1, D1, chi, device=device, dtype=dtype)

    einsum(T, U)
    '''
    the difference seems due to gpu lazy allocation or asynchronously ? (t1-t0 is always longer)
    https://discuss.pytorch.org/t/measuring-gpu-tensor-operation-speed/2513
    '''
    
    start = time.time()
    res1 = einsum(T, U)
    einsum_time = time.time()

    res2 = tensordot(T, U)
    tensordot_time = time.time()

    print (torch.dist(res1, res2).item(), einsum_time-start, tensordot_time-einsum_time)
